#ifndef __KERMIT_H__
#define __KERMIT_H__

/* kermit.h -- Symbol and struct definitions for embedded Kermit. */

/*
  Author: Frank da Cruz.
  Copyright (C) 1995, 2001, 
  Trustees of Columbia University in the City of New York.
  All rights reserved.
*/

#ifdef COMMENT                          /* COMMENT must not be defined */
#undef COMMENT
#endif /* COMMENT */

/*
  Never use NULL as a pointer.  Always use 0 cast to the appropriate type,
  for example: (UCHAR *)0.  Reason: system might define NULL to be an
  integer 0, and when passed to a function that wants a pointer, might wind
  up with junk in the left half.
*/
#ifdef NULL
#undef NULL
#endif /* NULL */

/* Feature Selection */

#define F_LP                            /* Long packets */
#define F_AT                            /* Attribute packets */
#define F_CTRLC                         /* 3 consecutive Ctrl-C's to quit */
#define F_SSW				/* Simulated sliding windows */
#define F_SCAN				/* Scan files for text/binary */
/*
  F_SSW means we say (in negotiations) that we support sliding windows, but we
  really don't.  This allows the sender to send to us in a steady stream, and
  works just fine except that error recovery is via go-back-to-n rather than
  selective repeat.
*/

#define NODEBUG

#ifndef NODEBUG
#define DEBUG
#endif /* NODEBUG */

#ifdef COMMENT                          /* None of the following ... */
/*
  + = It works if selected
  - = Partially implemented but doesn't work
  0 = Not implemented
*/
  #define DEBUG                         /* + Send debug output to stderr */
  #define F_TSW                         /* - True sliding windows */
  #define F_LS                          /* 0 Locking shifts */
  #define F_RS                          /* 0 Recovery */

#endif /* COMMENT */

#ifdef F_TSW				/* F_SW is defined if either */
#ifndef F_SW				/* F_SSW or F_TSW is defined... */
#define F_SW
#endif /* F_SW */
#endif /* F_TSW */

#ifdef F_SSW  
#ifndef F_SW
#define F_SW
#endif /* F_SW */
#endif /* F_SSW */

/* Control character symbols */

#define NUL  '\0'                       /* Null */
#define SOH  001                        /* Start of header */
#define LF   012                        /* Linefeed */
#define CR   015                        /* Carriage Return */
#define SO   016                        /* Shift Out */
#define SI   017                        /* Shift In */
#define DLE  020                        /* Datalink Escape */
#define ESC  033                        /* Escape */
#define XON  021                        /* XON */
#define XOFF 023                        /* XOFF */
#define SP   040                        /* Space */
#define DEL  0177                       /* Delete (Rubout) */

/*
typedef unsigned char UCHAR;
typedef unsigned long ULONG;
*/

/* Main program return codes */

#define SUCCESS     0
#define FAILURE     1

/* Tunable parameters */

#define OBUFLEN  1024                   /* File output buffer size */
#define IBUFLEN   512			/* File input buffer size */

#define FN_MAX     80                   /* Maximum filename length */
#define DATE_MAX   20                   /* Max length for file date */

#ifdef F_SW                             /* Window slots */
#ifdef F_TSW				/* True window slots */
#define P_WSLOTS    4			/* Max is 4 */
#else
#define P_WSLOTS   31			/* Simulated max is 31 */
#endif /* F_TSW */
#else
#define P_WSLOTS    1
#endif /* F_SW */

#ifdef F_LP                             /* Receive packet length */
#define P_PKTLEN 4096
#else
#define P_PKTLEN   90
#endif /* F_LP */

/* Generic On/Off values */

#define OFF         0
#define ON          1

/* File Transfer Modes */

#define BINARY      0
#define TEXT        1

/* Parity values */

#define PAR_NONE    0
#define PAR_SPACE   1
#define PAR_EVEN    2
#define PAR_ODD     3
#define PAR_MARK    4

/* Protocol parameters */

#define P_S_TIMO   40                   /* Timeout to tell other Kermit  */
#define P_R_TIMO    5                   /* Default timeout for me to use */
#define P_RETRY    10                   /* Per-packet retramsit limit    */
#define P_PARITY  PAR_NONE              /* Default parity        */
#define P_R_SOH   SOH                   /* Incoming packet start */
#define P_S_SOH   SOH                   /* Outbound packet start */
#define P_R_EOM    CR                   /* Incoming packet end   */
#define P_S_EOM    CR                   /* Outbound packet end   */

/* Capability bits */

#define CAP_LP      2                   /* Long packet capability */
#define CAP_SW      4                   /* Sliding windows capability */
#define CAP_AT      8                   /* Attribute packet capability */
#define CAP_RS     16                   /* Resend capability */
#define CAP_LS     32                   /* Locking shift capability */

/* Actions */

#define A_SEND      1			/* Send file(s) */
#define A_RECV      2			/* Receive file(s) */

/* Receive protocol states */

#define R_ERROR    -1                   /* Fatal protocol error */
#define R_NONE      0                   /* Protocol not running */
#define R_WAIT      1                   /* Waiting for S packet */
#define R_FILE      2                   /* Waiting for F or B packet */
#define R_ATTR      3                   /* Waiting for A or D packet */
#define R_DATA      4                   /* Waiting for D or Z packet */

/* Send protocol states */

#define S_ERROR    -1                   /* Fatal protocol error */
#define S_NONE     10                   /* Protocol not running */
#define S_INIT     11                   /* Sent S packet */
#define S_FILE     12                   /* Sent F packet */
#define S_ATTR     13                   /* Sent A packet */
#define S_DATA     14                   /* Sent D packet */
#define S_EOF      15                   /* Sent Z packet */
#define S_EOT      16                   /* Sent B packet */

/* What I'm Doing */

#define W_NOTHING   0
#define W_SEND      1
#define W_RECV      2

/* Kermit module function codes */

#define K_INIT      0                   /* Initialize */
#define K_RUN       1                   /* Run */
#define K_STATUS    2                   /* Request status */
#define K_CONTINUE  3                   /* Keep going */
#define K_QUIT      4                   /* Quit immediately */
#define K_ERROR     5                   /* Quit with error packet, msg given */
#define K_SEND      6			/* Begin Send sequence */

/* Need these? */

#define K_FILE      7			/* Send this file */
#define K_ATTR      8			/* Send file attributes */
#define K_EOT       9			/* End Send Sequence */

/* Kermit module return codes */

#define X_ERROR    -1                   /* Fatal error */
#define X_OK        0                   /* OK, no action needed */
#define X_FILE      1                   /* Filename received */
#define X_DATA      2                   /* File data received */
#define X_DONE      3                   /* Done */
#define X_STATUS    4                   /* Status report */

struct packet {
    int len;                            /* Length */
    short seq;                          /* Sequence number */
    char typ;                           /* Type */
    short rtr;                          /* Retry count */
    UCHAR * dat;                        /* Pointer to data */
    short flg;				/* Flags */
};

struct k_data {                         /* The Kermit data structure */
    short remote;			/* 0 = local, 1 = remote */
    short xfermode;			/* 0 = automatic, 1 = manual */
    short binary;                       /* 0 = text, 1 = binary */
    short state;                        /* Kermit protocol state */
    short what;				/* Action (send or receive) */
    short s_first;			/* Enocode at beginning of file */
    short s_next;			/* Encode lookahead byte */
    short s_seq;                        /* Sequence number sent */
    short r_seq;                        /* Sequence number received */
    short s_type;                       /* Packet type sent */
    short r_type;                       /* Packet type received */
    short s_soh;                        /* Packet start sent */
    short r_soh;                        /* Packet start received */
    short s_eom;                        /* Packet end sent */
    short r_eom;                        /* Packet end received */
    int size;				/* Current size of output pkt data */
    int osize;				/* Previous output packet data size */
    int r_timo;                         /* Receive and send timers */
    int s_timo;                         /* ... */
    int r_maxlen;                       /* maximum packet length to receive */
    int s_maxlen;                       /* maximum packet length to send */
    short window;                       /* maximum window slots */
    short wslots;                       /* current window slots */
    short parity;                       /* 0 = none, nonzero = some */
    short retry;                        /* retry limit */
    char s_ctlq;                        /* control-prefix out */
    char r_ctlq;                        /* control-prefix in */
    char ebq;				/* 8-bit prefix */
    char ebqflg;			/* 8-bit prefixing negotiated */
    char rptq;				/* Repeat-count prefix */
    int s_rpt;				/* Current repeat count */
    short rptflg;                       /* flag for repeat counts negotiated */
    short bct;                          /* Block-check type 1..3 */
    unsigned short capas;               /* Capability bits */
    long crcta[16];                     /* CRC generation table A */
    long crctb[16];                     /* CRC generation table B */
    UCHAR s_remain[6];			 /* Send data leftovers */
    UCHAR ipktbuf[P_PKTLEN+8][P_WSLOTS]; /* Buffers for incoming packets */
    struct packet ipktinfo[P_WSLOTS];    /* Incoming packet info */
#ifdef COMMENT
    UCHAR opktbuf[P_PKTLEN+8][P_WSLOTS]; /* Buffers for outbound packets */
#else
    UCHAR opktbuf[P_PKTLEN+8];		/* Outbound packet buffer */
    int opktlen;			/* Outbound packet length */
    UCHAR xdatabuf[P_PKTLEN+2];		/* Buffer for building data field */
#endif /* COMMENT */
    struct packet opktinfo[P_WSLOTS];	/* Outbound packet info */
    UCHAR * xdata;			/* Pointer to data field of outpkt */
#ifdef F_TSW
    short r_pw[64];			/* Packet Seq.No. to window-slot map */
    short s_pw[64];			/* Packet Seq.No. to window-slot map */
#endif /* F_TSW */
    UCHAR ack_s[32];                    /* Our own init parameter string */
    UCHAR * obuf;
    int obuflen;                        /* Length of output file buffer */
    int obufpos;                        /* Output file buffer position */
    UCHAR ** filelist;			/* List of files to send */
    UCHAR * dir;			/* Directory */
    UCHAR * sendfile;			/* Name of current file */
    UCHAR * istring;			/* Pointer to string to encode from */
    UCHAR * ostring;			/* Pointer to string to decode to */
    int (*txd)(UCHAR *, int, short);    /* pointer to comms write function */
    int (*openf)(UCHAR *, int, struct k_data *); /* file-open func pointer */
    int (*readf)(struct k_data *);	/* pointer to file-read  function */
    int (*writef)(struct k_data *, UCHAR *, int);        /* pointer to file-write function */
    int (*closef)(UCHAR,int);		/* pointer to file-close function */
    UCHAR zinbuf[IBUFLEN+1];		/* Pointer to input file buffer */
    UCHAR * zinptr;
    int zincnt;				/* Length of input file buffer */
};

struct k_response {
    short status;                       /* From Kermit: current status */
    UCHAR filename[FN_MAX];             /* From Kermit: name of file */
    UCHAR filedate[DATE_MAX];           /* From Kermit: date of file */
    long filesize;                      /* From Kermit: size of file */
};

/* Macro definitions */

#define tochar(ch)  (UCHAR)((UCHAR)((UCHAR)(ch) + SP ))
#define xunchar(ch) (UCHAR)((UCHAR)((UCHAR)(ch) - SP ))
#define ctl(ch)     (UCHAR)((UCHAR)((UCHAR)(ch) ^ 64 ))

#ifdef COMMENT
#define tochar(ch)  (((ch) + SP ) & 0xFF ) /* Digit to character */
#define xunchar(ch) (((ch) - SP ) & 0xFF ) /* Character to number */
#define ctl(ch)     (((ch) ^ 64 ) & 0xFF ) /* Controllify/uncontrollify */
#endif /* COMMENT */

/* Prototypes for i/o functions */

int devopen(char *);
int pktmode(short);
int readpkt(UCHAR *, struct k_data *);
int inchk(void);
int tx_data(UCHAR *, int, short);
int devclose(void);
int openfile(UCHAR *, int, struct k_data *);
ULONG fileinfo(UCHAR *, UCHAR *, int, short *, short);
int writefile(struct k_data *, UCHAR *, int);
int readfile(struct k_data *);
int closefile(UCHAR, int);

/* Prototypes for kermit() functions */

int kermit(short, struct k_data *, short, int, char *, struct k_response *);
UCHAR * getrslot(struct k_data *, short *);
UCHAR * getsslot(struct k_data *, short *);
void freerslot(struct k_data *, short);
void freesslot(struct k_data *, short);
int spkt(char, short, int, UCHAR *, struct k_data *);
int ack(struct k_data *, short, UCHAR * text);
int nak(struct k_data *, short, short);
int chk1(UCHAR *);
unsigned int chk2(UCHAR *);
unsigned int chk3(UCHAR *, struct k_data *);
void spar(struct k_data *, UCHAR *, int);
int rpar(struct k_data *, char);
int decode(struct k_data *, struct k_response *, short, UCHAR *);
ULONG stringnum(UCHAR *);
UCHAR * numstring(ULONG, UCHAR *, int);
void gattr(UCHAR *, struct k_response *);
int sattr(struct k_data *);
int sdata(struct k_data *);
void epkt(char *, struct k_data *);
int encstr(UCHAR *, struct k_data *);
void decstr(UCHAR *, struct k_data *, struct k_response *);
void encode(int, int, struct k_data *);
int nxtpkt(struct k_data *);
int resend(struct k_data *);
#ifdef DEBUG
int xerror(void);
#endif /* DEBUG */

#endif /* End of kermit.h */
